#include <stdio.h>
#include <string.h>
#include "occ.h"
#include "ex1.h"
void PrintVec(double *v, int n, const char *nm);
void PrintVec_C(double *v, int n, const char *nm);
void PrintArray2D(Array2D &A, const char *nm);
void PrintArray2DT(Array2D &A, const char *nm);
void PrintArray2D_C(Array2D &A, const char *nm);
void PrintXY(double *x, double *y, double *y2, int n);
static double Thparm, Aparm;
static int method;
static FILE *fout;
// =================================================================
void SolveBVP::Solve(int nt){
   int s, t, g;
	int i,j;
	double *rc;
   const int nflux = 7;
   const int nint = 21;
   double xi[nint],del;
	Array2D yi(5,nint);
	Array2D flux(4,nflux);
   Orthcc cc;
// ----------------------------------------------
	n = nt-2;
// This program solves diffusion/conduction with linear source term
// having a spatially varying source strength
   // allocate memory ---------------------------
   xc = new double[nt]; // quadrature base points
   wc = new double[nt]; // quadrature weights
   rc = new double[nt]; // rate constant at xc
	Array2D &A  = *(new Array2D(nt,nt+1));
	Array2D &Aa  = *(new Array2D(n,n));
	Array2D &yc  = *(new Array2D(5,nt));
   // set up & get arrays -----------------------
	Thx = 4.0*Thparm*Thparm;
   xi[0]=0.0; yi.v[0][0] = 0.0;
   del = 1.0/double(nint-1);
   for(i=1; i<nint; i++){
      xi[i] = xi[i-1] + del;
      yi.v[0][i] = xi[i];
   }
   s = cc.Nonsymmetric;
   t = method;
   g = cc.Planar;
   cc.occ_set(s,t,g);
   cc.Quadrature(xc,wc,n);
   cc.Acoeff(Ac,xc,n);           // first derivative
   cc.Bcoeff(Bc,Ac,xc);          // 2nd derivative
   cc.Ccoeff(Cc,Ac,xc,wc);       // symmetric 2nd derivative
   cc.MassMat(Dc,xc,wc,n);       // mass matrix
   cc.LcoeffN(Lc,xc,xi,nt,nint); // interpolating polynomials
	ReportBasic();    // basic report
   FuncRX(xc,rc,nt); // rate parameters
   for(i=0; i<nt; i++)yc.v[0][i] = xc[i];  // save for output
   // -----------------------------------------------------
   // 1. Solution by standard collocation
   OCC_Solve(A,Aa,rc,yc.v[1]);
   Interp_y(yc.v[1],yi.v[1],nint,Lc);
   FluxCalc(0,flux.v[0],yc.v[1],rc,wc,A,Ac);
   // -----------------------------------------------------
   // 2. Solution by symmetric matrix problem
   OCC_Sym_Solve(A,Aa,rc,yc.v[2]);
   Interp_y(yc.v[2],yi.v[2],nint,Lc);
   FluxCalc(1,flux.v[1],yc.v[2],rc,wc,A,Ac);
   // -----------------------------------------------------
   // 3. Galerkin/Moments with interpolated rates
   MM_interp_Solve(A,Aa,rc,yc.v[3]);
   Interp_y(yc.v[3],yi.v[3],nint,Lc);
   FluxCalc(2,flux.v[2],yc.v[3],rc,wc,A,Ac);
   // -----------------------------------------------------
   // 4. Galerkin/Moments with exact integration
   MM_Full_Solve(A,Aa,yc.v[4]);
   Interp_y(yc.v[4],yi.v[4],nint,Lc);
   FluxCalc(3,flux.v[3],yc.v[4],rc,wc,A,Ac);
   // Result Summary
   ReportSolution(yc, yi, flux);
   // free memory -------------------------------
	delete [] xc; 	delete [] wc; 	delete [] rc;
	delete &A; 	delete &Aa;  delete &yc;
   return;
}
//===================================================================================
void SolveBVP::OCC_Solve(Array2D &A, Array2D &Aa, double rc[], double yc[]){
   int i,j,nt;
	int status;
// --------------------------------------------------
// Solve using conventional Orthogonal Collocatoin
// --------------------------------------------------
   nt = n + 2;
	for(j=0; j<nt; j++){
	   for(i=0; i<nt; i++){
	      A.v[i][j] = Bc.v[i][j];
	   }
	   A.v[j][j] -= rc[j];  // added to diagonal
	   A.v[j][nt] = -rc[j];  // rhs
	}
   CopyAA(A,Aa,yc);              // copy
 	status = Aa.LUSolveA(&yc[1]); // solve not symmetric
	yc[0]=0.0; 	yc[n+1] = 0.0;
   PrintArray2D(A,"Conventional collocation - nonsymmetric matrix problem:");
   PrintVec(yc,nt,"Conventional collocation solution:");
   PrintVec_C(yc,nt,"Conventional collocation solution:");
}
//===================================================================================
void SolveBVP::OCC_Sym_Solve(Array2D &A, Array2D &Aa, double rc[], double yc[]){
   int i,j,nt;
	int status;
// --------------------------------------------------
// Solve using Orthogonal Collocation with symmetric matrix
// --------------------------------------------------
   nt = n + 2;
	for(j=0; j<nt; j++){
	   for(i=0; i<nt; i++){
	      A.v[i][j] = Cc.v[i][j];
	   }
	   A.v[j][j] += wc[j]*rc[j];
	   A.v[j][nt] = wc[j]*rc[j];
	}
   CopyAA(A,Aa,yc);  // save copy
 	status = Aa.LUSolveA(&yc[1]);  // does not exploit symmetry
	yc[0]=0.0;	yc[nt-1] = 0.0;
   PrintArray2D(A,"Alternate form - symmetric matrix problem:");
   PrintVec(yc,nt,"Symmetric form solution:");
   PrintVec_C(yc,nt,"Symmetric form solution:");
}
//===================================================================================
void SolveBVP::MM_interp_Solve(Array2D &A, Array2D &Aa, double rc[], double yc[]){
   int i,j,nt;
	int status;
   double s;
// --------------------------------------------------
// Solve using Galerkin/Moments with interpolated rates
// --------------------------------------------------
   nt = n + 2;
	for(i=0; i<nt; i++){
      s = 0.0;
	   for(j=0; j<nt; j++){
         Dc.v[i][j] *= rc[j];
         s += Dc.v[i][j];
	      A.v[i][j] = Cc.v[i][j] + Dc.v[i][j];
	   }
	   A.v[i][nt] = s;   // rhs
	}
   CopyAA(A,Aa,yc);  // save copy
 	status = Aa.LUSolveA(&yc[1]);  // not symmetric
	yc[0]=0.0;	yc[nt-1] = 0.0;
   PrintArray2D(A,"Galerkin/Moments with interpolated rates:");
   PrintVec(yc,nt,"Solution interpolated rates:");
   PrintVec_C(yc,nt,"Solution interpolated rates:");
}

//===================================================================================
void SolveBVP::MM_Full_Solve(Array2D &A, Array2D &Aa, double yc[]){
   int i,j,nt;
	int status;
// --------------------------------------------------
// Solve using Galerkin/Moments exact mass matrix
// --------------------------------------------------
   nt = n + 2;
   Array2D &D = *(new Array2D(nt,nt));
   MMatrix(D);
	for(i=0; i<nt; i++){
      double s = 0.0;
	   for(j=0; j<nt; j++){
         s += D.v[i][j];
	      A.v[i][j] = Cc.v[i][j] + D.v[i][j];
	   }
      A.v[i][nt] = s;
	}
   CopyAA(A,Aa,yc);  // save copy
 	status = Aa.LUSolveA(&yc[1]);  // Galerkin Symmetric not exploited
	yc[0]=0.0;	yc[nt-1] = 0.0;
   PrintArray2D(D,"Galerkin/Moments exact mass matrix:");
   PrintArray2D(A,"Galerkin/Moments with exact integration:");
   PrintVec(yc,nt,"Exact mass matrix solution:");
   PrintVec_C(yc,nt,"Exact mass matrix solution:");
   delete &D;
}
//===================================================================================
void SolveBVP::FluxCalc(int meth, double flux[], double yc[], double rc[], double wc[], Array2D &A, Array2D &Ac){
   int i,nt;
   double s0,s1,a0,a1,ra;
   // flux as follows: -------------------
   //  0 - overall calculated from integral
   //  1 - x = 0, left side correct
   //  2 - x = 1, right side correct
   //  3 - overall by sum of 1 and 2
   //  4 - first derivative at x = 0
   //  5 - negative of first derivative at x = 1
   //  6 - overall by sum of 4 and 5
   // Note: 1 & 2 are calculated from solution matrix evaluated at 0 and n+1
   // i.e. Eq. (1.6) for conventional orthogonal collocation using B matrix
   // or using Eq. (1.13) for other formulations
   // A is the matrix problem, rhs stored at column n+2
   // Ac is the first derivative operator
   // -------------------------------------
   nt = n + 2;
   a0 = 0.0;  a1 = 0.0, ra = 0;
   s0 = A.v[0][nt] ; s1 = A.v[n+1][nt];  // RHS see solve routines
	for(i=0; i<nt; i++){
      ra += wc[i]*rc[i]*(1.0 - yc[i]);
	   s0 -= A.v[0][i]*yc[i];     // left eqn. at j = 0
	   s1 -= A.v[n+1][i]*yc[i];   // right eqn. at j = n+1
	   a0 += Ac.v[0][i]*yc[i];    // left first derivative
	   a1 -= Ac.v[n+1][i]*yc[i];  // right first derivative
   }
   ra /= Thx; s0 /= Thx; s1 /= Thx; a0 /= Thx; a1 /= Thx;
   if(meth == 0){             // conventional s0 & s1 are boundary residuals
      s0 = a0 - wc[0]*s0;     // apply Eq. (1.24)
      s1 = a1 - wc[n+1]*s1;
   }
   flux[0] = ra;
   flux[1] = s0; flux[2] = s1; flux[3] = s0+s1;
   flux[4] = a0; flux[5] = a1; flux[6] = a0+a1;
}
//===================================================================================
void SolveBVP::Interp_y(double yc[], double yi[], int ni, Array2D &L){
   int nt,i,j;
   double s;
   // Fill in between points using interpolation
   nt = n + 2;
	for(i=0; i<ni; i++){
      s = 0.0;
	   for(j=0; j<nt; j++){
         s += L.v[i][j]*yc[j];
      }
      yi[i] = s;
   }
}
//===================================================================================
void SolveBVP::MMatrix(Array2D &D){
   int i,j,k;
   Orthcc cc;
   // Calculate mass matrix array using Gaussian quadrature
   //  Method   weight func  trial func  reaction total Gauss Reqd
   //  Moments       n-1         n+1        3     2n+3    m = n+2
   //  Galerkin      n+1         n+1        3     2n+5    m = n+3
   int nt = n+2;
   int m = n+3;                  // Gauss points for Galerkin
   if(method == cc.Gauss)m--;    // n+2, need one less quadrature point
   double *x = new double[m+2];  // quadrature base points
   double *w = new double[m+2];  // quadrature weights
   double *r = new double[m];    // rate terms
   Array2D &L = *(new Array2D(m,n+2));
   // ---------------------------------------------------
   cc.occ_set(cc.Nonsymmetric,cc.Gauss,cc.Planar);
   cc.Quadrature(x,w,m);  // first and last w are 0
   for(k=0;k<m;k++){
      x[k] = x[k+1]; w[k] = w[k+1];
   }
   FuncRX(x,r,m);
   for(k=0; k<m; k++){
      w[k] *= r[k];
      cc.Lcoeff(L.v[k],xc,x[k],n+2);
   }
   if(method == cc.Gauss){ // Moments only
      for(k=0;k<m;k++)
         w[k] /= (x[k]*(1.0 - x[k]));
   }
   for(i=0;i<nt;i++){
      for(j=i;j<nt;j++){
         double s = 0.0;
         for(k=0;k<m;k++){
            s += w[k]*L.v[k][i]*L.v[k][j];
         }
         D.v[i][j] = s;
         D.v[j][i] = s;
      }
   }
   if(method == cc.Gauss){ // Moments only
      for(i=1;i<nt-1;i++){
         for(j=0;j<nt;j++)
            D.v[i][j] *= xc[i]*(1.0 - xc[i]);
      }
      for(j=0;j<nt;j++){
         D.v[0][j] = 0.0;  D.v[nt-1][j] = 0.0;
      }
   }  // endif Gauss
   // --------- cleanup & reset ---------------------------
   cc.occ_set(cc.Nonsymmetric,method,cc.Planar);
	delete [] x;
	delete [] w;
   delete [] r;
   delete &L;
}
//===================================================================================
void SolveBVP::CopyAA(Array2D &A, Array2D &Aa, double *y){
   // A is includes boundary points, rhs in last column
   // Aa is the interior parts
	for(int i=0; i<n; i++){
	   for(int j=0; j<n; j++){
		  Aa.v[i][j] = A.v[i+1][j+1];
       }
       y[i+1] = A.v[i+1][n+2];  // rhs
	}
}
//===================================================================================
void SolveBVP::FuncRX(double *x, double *r, int m){
   int i;
	double Aparm1 = (1.0 - Aparm)*4.0;
   for(i=0; i<m; i++){
      r[i] = Thx*(Aparm + Aparm1*(1.5 - x[i])*x[i]*x[i]);
   }
}
//===================================================================================
void SolveBVP::ReportBasic(){
   const char *smeth[] = {"Gauss/Moments", "Lobatto/Galerkin"};
   fprintf(fout,"\nSolution using %s with %i interior points\n",smeth[method-1],n);
	fprintf(fout,"Thiele Modulus    :\t%8.2f\n",Thparm);
	fprintf(fout,"Reaction variation:\t%8.4f\n",Aparm);
   PrintVec(xc,n+2,"X - base points");
   PrintVec(wc,n+2,"W - quadrature weights");
   PrintArray2D(Ac,"A - first derivative array:");
   PrintArray2D(Bc,"B - Laplacian:");
   PrintArray2D(Cc,"C Matrix (AtWA or -WB):");
   PrintArray2D(Dc,"Mass Matrix:");
}
//===================================================================================
void SolveBVP::ReportSolution(Array2D &yc, Array2D &yi, Array2D &flux){
   PrintArray2DT(yc,"Solutions:");
   PrintArray2DT(yi,"Interpolated Solutions:");
   PrintArray2D(flux,"Flux Calculations:");
}
//===================================================================================
void PrintVec(double *v, int n, const char *nm){
   int i;
	fprintf(fout,"%s\n",nm);
   for(i=0; i<n; i++)fprintf(fout,"%14.8f\t",v[i]);
   fprintf(fout,"\n");
}
//===================================================================================
void PrintVec_C(double *v, int n, const char *nm){
   int i;
	printf("%s\n",nm);
   for(i=0; i<n; i++)printf("%14.8f\t",v[i]);
   printf("\n");
}
//===================================================================================
void PrintArray2D(Array2D &A, const char *nm){
   int i,j;
	fprintf(fout,"%s\n",nm);
	for(i=0; i<A.n0; i++){
	   for(j=0; j<A.n1; j++)fprintf(fout,"%14.8f\t",A.v[i][j]);
      fprintf(fout,"\n");
	}
}
//===================================================================================
void PrintArray2D_C(Array2D &A, const char *nm){
   int i,j;
	printf("%s\n",nm);
	for(i=0; i<A.n0; i++){
	   for(j=0; j<A.n1; j++)printf("%14.8f\t",A.v[i][j]);
      printf("\n");
	}
}
//===================================================================================
void PrintArray2DT(Array2D &A, const char *nm){   // reverse rows and columns
   int i,j;
	fprintf(fout,"%s\n",nm);
	for(j=0; j<A.n1; j++){
	   for(i=0; i<A.n0; i++)fprintf(fout,"%14.8f\t",A.v[i][j]);
      fprintf(fout,"\n");
	}
}
//===================================================================================
void PrintXY(double *x, double *y, double *y2, int n){
   int i;
   fprintf(fout,"Interpolated Solution:\n");
   for(i=0; i<n; i++){
      fprintf(fout,"\t%10.8f\t%10.8f\t%10.8f\n",x[i],y[i],y2[i]);
   }
}
int main(int argc, char* argv[]){
	int nt, n0,n1,ndel;
	static const int LINE_LEN = 128;
	static char line[LINE_LEN];
   Orthcc cc;     // class constructor
	SolveBVP Ex;   // constructor
	char fnm[LINE_LEN];
/* 
	This console application solves a nonsymmetric reaction and diffusion problem.
	Either Gauss/Moments or Lobatto/Galerkin methods may be selected.
	The problem is solved for the specified range of nodal points.
	Thparm - is the Thiele Modulus
	Aparm - 0 < Aparm <= 1.0, governs variation of reactivity with position. The reactivity is
			Aparm at x =0 and 2 - Aparm at x=1. The average value is 1.0
	The output file is written in tab delimited text for easy import to a apreadsheet.
*/
   char chs = 'c', crx;
	printf(" Enter Thiele Modulus & variation parameter:\n");
	scanf("%lf %lf%c",&Thparm,&Aparm,&crx);
   printf("\nEnter g/l for Gauss or Lobatto points >");
   scanf("%c%c",&chs,&crx);
   method = cc.Lobatto;
   if(chs == 'g')method = cc.Gauss;
   printf("\nEnter interior points, 1st, last, delta >");
   scanf("%i %i %i%c",&n0,&n1,&ndel,&crx);
   printf("\nEnter root of output file >");
   scanf("%s%c",fnm,&crx);
   strcat(fnm,".dat");
   fout = fopen(fnm,"w");
	for(nt=n0+2; nt<=n1+2; nt+=ndel){
      Ex.Solve(nt);
	}
	return 0;
}
