#include <stdio.h>
#include <string.h>
#include <math.h>
#include "occ.h"
void PrintVec(double *v, int n, const char*);
void PrintArray2D(Array2D &A, const char*);
void PrintArrayX(double *v, Array2D &A, const char *);
FILE *fout =  stdout;
int main( ){
   int n,ntot,i,j;
   char fnm[80];
   const char *symmetry[2] = {"nonsymmetric","symmetric"};
   const char *method[2] = {"Gauss", "Lobatto"};
   const char *geometry[3] = {"planar", "cylindrical", "spherical"};
   // set up equal spaced interpolation points
   const int nint=11;
   double xi[nint], del = 1.0/double(nint-1);
   xi[0]=0.0;
   for(i=1; i<nint; i++)xi[i] = xi[i-1] + del;
   // arrays for collocation -----------------------------
   double *x, *w; // points and quadrature weights
   Array2D A;     // first derivative
   Array2D An;    // first derivative odd function
   Array2D B;     // second derivative
   Array2D C;     // second derivative, symmetric matrix
   Array2D D;     // mass matrix
   Orthcc cc(0,2,0);  // class constructor
   double *ex;
   // get basic data -------------------------------------
   // set symmetry, type (Gauss or Lobatto), geometry (if symmetric)
   int n0 = 3;  // interior points, first
   int n1 = 3;  // interior points, last
   int s = cc.Nonsymmetric;   // s is symmetry
   int t = cc.Lobatto;        // t is Gauss or Lobatto
   int g = cc.Cylindrical;    // g is geometry
   char chs = 'c', crx;
   printf("\nEnter y/n for symmetric/nonsymmetric  >");
   scanf("%c%c",&chs,&crx);  // crx for carriage return
   if(chs == 'y')s = cc.Symmetric;
   printf("\nEnter g/l for Gauss or Lobatto points >");
   scanf("%c%c",&chs,&crx);
   if(chs == 'g')t = cc.Gauss;
   if(s == cc.Symmetric){
      printf("\nEnter 0/1/2 for planar/cylindrical/spherical >");
      scanf("%i%c",&g,&crx);
      g = (g < 0) ? 0 : g;
      g = (g > 2) ? 2 : g;
   }
   printf("\nEnter interior points, 1st and last >");
   scanf("%i %i%c",&n0,&n1,&crx);
   //printf("\nResults: %i %i %i\n",n0,n1,crx);
   printf("\nEnter root of output file >");
   scanf("%s%c",fnm,&crx);
   strcat(fnm,".dat");
   fout = fopen(fnm,"w");
   // -------------------------------------
   cc.occ_set(s,t,g);  // set up type of problem
   printf("\nOrthogonal collocation arrays for %s problems with %s geometry\n",symmetry[s],geometry[g]);
   fprintf(fout,"Orthogonal collocation arrays for %s problems with %s geometry\n",symmetry[s],geometry[g]);
   for(n=n0; n<=n1; n++){
      printf("The number of interior %s points n = %i\n",method[t-1],n);
      fprintf(fout,"The number of interior %s points n = %i\n",method[t-1],n);
      ntot = n + 2 - s;  // total points
      // allocate memory (Array class care of others)
      x = new double[ntot];
      w = new double[ntot];
      ex = new double[ntot];
      Array2D &L = *(new Array2D(nint,ntot));
      Array2D &Ex = *(new Array2D(nint,2));  // interpolation test
      // get arrays
      cc.Quadrature(x,w,n);
      cc.Acoeff(A,x,n);
      cc.Anonsym(An,A,x);
      cc.Bcoeff(B,A,x);
      cc.Ccoeff(C,A,x,w);
      cc.MassMat(D,x,w,n);
      for(i=0; i<nint; i++)
        cc.Lcoeff(L.v[i],x,xi[i],ntot);
      // print arrays -------------------------------------
      PrintVec(x,ntot,"Base Points:");
      PrintVec(w,ntot,"Quadrature Weights:");
      // ------ first derivatives
      PrintArray2D(A,"A - first derivative array:");
      if(s == cc.Symmetric){
         PrintArray2D(An,"A* - first derivative of odd function");
      }
      PrintArray2D(B,"B - Laplacian:");
      PrintArray2D(C,"C symmetric Laplacian (AtWA or -WB):");
      PrintArray2D(D,"D Mass Matrix:");
      PrintArrayX(xi,L,"Interpolating Polynomial Values:");
      // test interpolation and integration ---------------
      double sum = 0.0, aexp = 3.0;
      double sint =(exp(aexp) - 1)/(2.0*aexp);
      for(i=0;i<ntot;i++){
         ex[i] = exp(aexp*x[i]*x[i]);
         sum += w[i]*ex[i];
      }
      for(j=0; j<nint; j++){
         Ex.v[j][0] = exp(aexp*xi[j]*xi[j]);
         Ex.v[j][1] = 0.0;
         for(i=0;i<ntot;i++){
            Ex.v[j][1] += ex[i]*L.v[j][i];
         }
      }
      fprintf(fout,"Exact and approximate integral exp(a*x^2)\t%11.7f\t%11.7f\t%10.4E\n",
         sint,sum,sint-sum);
      PrintArrayX(xi,Ex,"Interpolation of exp(a*x^2):");
      // free space
      delete [] x;
      delete [] w;
      delete &L;
      delete &Ex;
      delete [] ex;
   }
   // -----------------------------
   return 0;
}
void PrintVec(double *v,int n, const char *title){
   fprintf(fout,"%s\n",title);
   for(int i=0; i<n; i++)fprintf(fout,"%10.8f\t",v[i]);
   fprintf(fout,"\n");
}
void PrintArray2D(Array2D &A, const char *title){
   fprintf(fout,"%s\n",title);
	for(int i=0; i<A.n0; i++){
	   for(int j=0; j<A.n1; j++)fprintf(fout,"%13.6E\t",A.v[i][j]);
      fprintf(fout,"\n");
	}
}
void PrintArrayX(double *v, Array2D &A, const char *title){
   fprintf(fout,"%s\n",title);
	for(int i=0; i<A.n0; i++){
      fprintf(fout,"%13.6E\t",v[i]);
	   for(int j=0; j<A.n1; j++)fprintf(fout,"%13.6E\t",A.v[i][j]);
      fprintf(fout,"\n");
	}
}

